"""
Implements a basic command-line interface.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import argparse
import sys

from .core import check_url, sample_urls


def parse_args(args):
    """Define parser for command-line arguments"""
    argsparser = argparse.ArgumentParser(description='Command-line interface for Courlan')
    group1 = argsparser.add_argument_group('I/O', 'Manage input and output')
    group1.add_argument("-i", "--inputfile",
                            help='name of input file (required)',
                            type=str, required=True)
    group1.add_argument("-o", "--outputfile",
                            help='name of output file (required)',
                            type=str, required=True)
    group1.add_argument("-d", "--discardedfile",
                            help='name of file to store discarded URLs (optional)',
                            type=str)
    group1.add_argument("-v", "--verbose",
                            help="increase output verbosity",
                            action="store_true")
    group2 = argsparser.add_argument_group('Filtering', 'Configure URL filters')
    group2.add_argument("--strict",
                            help="perform more restrictive tests",
                            action="store_true")
    group2.add_argument("-l", "--language",
                            help="use language filter (ISO 639-1 code)",
                            type=str)
    group2.add_argument("-r", "--redirects",
                            help="check redirects",
                            action="store_true")
    group3 = argsparser.add_argument_group('Sampling', 'Use sampling by host, configure sample size')
    group3.add_argument('--sample',
                            help='use sampling',
                            action="store_true")
    group3.add_argument('--samplesize',
                            help='size of sample per domain',
                            type=int, default=1000)
    group3.add_argument('--exclude-max',
                            help='exclude domains with more than n URLs',
                            type=int) # default=10000
    group3.add_argument('--exclude-min',
                            help='exclude domains with less than n URLs',
                            type=int)
    return argsparser.parse_args()


def process_args(args):
    """Start processing according to the arguments"""
    if args.sample is False:
        with open(args.inputfile, 'r', encoding='utf-8', errors='ignore') as inputfh, \
             open(args.outputfile, 'w', encoding='utf-8') as outputfh:
            for line in inputfh:
                result = check_url(line, strict=args.strict, with_redirects=args.redirects, language=args.language)
                if result is not None:
                    outputfh.write(result[0] + '\n')
                # proceed with discarded URLs. to be rewritten
                elif args.discardedfile is not None:
                    with open(args.discardedfile, 'a', encoding='utf-8') as discardfh:
                        discardfh.write(line)
    else:
        urllist = []
        with open(args.inputfile, 'r', encoding='utf-8', errors='ignore') as inputfh:
            urllist.extend(line.strip() for line in inputfh)
        with open(args.outputfile, 'w', encoding='utf-8') as outputfh:
            for url in sample_urls(urllist, args.samplesize, exclude_min=args.exclude_min, exclude_max=args.exclude_max, strict=args.strict, verbose=args.verbose):
                outputfh.write(url + '\n')


def main():
    """Run as a command-line utility."""
    args = parse_args(sys.argv[1:])
    process_args(args)


if __name__ == '__main__':
    main()



"""
Unit tests for the UrlStore class of the courlan package.
"""

import os
import pickle
import signal
import uuid

from datetime import datetime

import pytest

from courlan import UrlStore


def test_urlstore():
    'Test all functionality related to the class.'

    # sanity checks
    my_urls = UrlStore()
    candidates = ['123', 'http://', 'ftp://example.org', 'http://example.org/']
    my_urls.add_urls(candidates)
    assert len(my_urls.urldict) == 1
    firstelem = my_urls.urldict['http://example.org'].tuples[0]
    assert firstelem.urlpath == '/' and firstelem.visited is False
    my_urls.urldict['http://example.org'].rules = pickle.loads(b'\x80\x03curllib.robotparser\nRobotFileParser\nq\x00)\x81q\x01}q\x02(X\x07\x00\x00\x00entriesq\x03]q\x04X\r\x00\x00\x00default_entryq\x05NX\x0c\x00\x00\x00disallow_allq\x06\x89X\t\x00\x00\x00allow_allq\x07\x89X\x03\x00\x00\x00urlq\x08X\x1f\x00\x00\x00https://sitemaps.org/robots.txtq\tX\x04\x00\x00\x00hostq\nX\x0c\x00\x00\x00sitemaps.orgq\x0bX\x04\x00\x00\x00pathq\x0cX\x0b\x00\x00\x00/robots.txtq\rX\x0c\x00\x00\x00last_checkedq\x0eGA\xd8\x87\xf5\xdc\xab\xd5\x00ub.')
    assert my_urls.urldict['http://example.org'].rules is not None
    assert my_urls.get_rules('http://example.org') == my_urls.urldict['http://example.org'].rules

    # filters
    my_urls = UrlStore(language='en', strict=True)
    candidates = ['https://de.wikipedia.org/', 'https://www.sitemaps.org/en_GB/protocol.html', 'http://example.com/de/test.html']
    my_urls.add_urls(candidates)
    assert len(my_urls.urldict) == 1 and 'https://www.sitemaps.org' in my_urls.urldict
    my_urls.add_urls(['https://www.sitemaps.org/es/1'], appendleft=['https://www.sitemaps.org/fi/2'])
    assert len(my_urls.urldict['https://www.sitemaps.org'].tuples) == 1

    # try example URLs
    example_domain = 'https://www.example.org'
    example_urls = [f'{example_domain}/{str(a)}' for a in range(10000)]
    test_urls = [
        f'https://test.org/{str(uuid.uuid4())[:20]}' for _ in range(10000)
    ]
    urls = example_urls + test_urls

    # compression 1
    my_urls = UrlStore(compressed=True)
    url_buffer = UrlStore()._buffer_urls(example_urls)
    my_urls.add_urls(example_urls)
    # necessary to pickle
    my_urls._lock = None
    assert len(pickle.dumps(my_urls)) < len(pickle.dumps(url_buffer))
    assert my_urls.is_known(f'{example_domain}/100') is True
    # compression 2
    my_urls = UrlStore(compressed=True)
    url_buffer = UrlStore()._buffer_urls(test_urls)
    my_urls.add_urls(test_urls)
    # necessary to pickle
    my_urls._lock = None
    assert len(pickle.dumps(my_urls)) < len(pickle.dumps(url_buffer))

    # test loading
    url_buffer = UrlStore()._buffer_urls(urls)
    assert sum(len(v) for _, v in url_buffer.items()) == len(urls)
    my_urls = UrlStore()
    my_urls.add_urls(urls)
    assert sum(
        len(my_urls._load_urls(k)) for k, _ in my_urls.urldict.items()
    ) == len(urls)
    assert my_urls.total_url_number() == len(urls)

    if my_urls.compressed is False:
        assert sum(len(v.tuples) for _, v in my_urls.urldict.items()) == len(urls)
    my_urls.add_urls(['https://visited.com/visited'], visited=True)
    assert my_urls.urldict['https://visited.com'].tuples[0].visited is True
    assert my_urls.urldict['https://visited.com'].all_visited is True
    assert my_urls.is_exhausted_domain('https://visited.com') is True
    # new unvisited URLs
    my_urls.add_urls(['https://visited.com/1'], visited=False)
    assert my_urls.urldict['https://visited.com'].tuples[1].visited is False
    assert my_urls.urldict['https://visited.com'].all_visited is False
    assert my_urls.is_exhausted_domain('https://visited.com') is False
    with pytest.raises(KeyError):
        assert my_urls.is_exhausted_domain('https://visited2.com') is True
    # revert changes for further tests
    del my_urls.urldict['https://visited.com'].tuples[1]
    my_urls.urldict['https://visited.com'].all_visited = True

    # test extension
    extension_urls = [f'{example_domain}/1/{str(a)}' for a in range(10)]
    my_urls.add_urls(extension_urls)
    assert len(my_urls._load_urls(example_domain)) == len(example_urls) + 10
    # test extension + deduplication
    extension_urls = [f'{example_domain}/1/{str(a)}/' for a in range(11)]
    my_urls.add_urls(appendleft=extension_urls)
    url_tuples = my_urls._load_urls(example_domain)
    assert len(url_tuples) == len(example_urls) + 11
    assert url_tuples[-1].urlpath == '/1/9' and url_tuples[0].urlpath == '/1/10/'

    # duplicates
    my_urls.add_urls(extension_urls)
    my_urls.add_urls(appendleft=extension_urls)
    assert len(my_urls._load_urls(example_domain)) == len(example_urls) + len(extension_urls)
    assert url_tuples[-1].urlpath == '/1/9' and url_tuples[0].urlpath == '/1/10/'

    # get_url
    assert my_urls.urldict[example_domain].timestamp is None
    assert my_urls.urldict[example_domain].count == 0
    url1 = my_urls.get_url(example_domain)
    url2 = my_urls.get_url(example_domain)
    assert url1 != url2 and url1 == 'https://www.example.org/1/10/'
    assert my_urls.urldict[example_domain].count == 2
    url_tuples = my_urls._load_urls(example_domain)
    # positions
    assert url1.endswith(url_tuples[0].urlpath) and url2.endswith(url_tuples[1].urlpath)
    # timestamp
    assert my_urls.urldict[example_domain].timestamp is not None
    # nothing left
    assert my_urls.urldict[example_domain].all_visited is False
    my_urls.add_urls(['http://tovisit.com/page'])
    assert my_urls.get_url('http://tovisit.com') == 'http://tovisit.com/page'
    assert my_urls.urldict['http://tovisit.com'].all_visited is True
    assert my_urls.get_url('http://tovisit.com') is None

    # known or not
    assert my_urls.is_known('http://tovisit.com/page') is True
    assert my_urls.is_known('https://www.other.org/1') is False
    assert my_urls.is_known('https://www.example.org/1') is True
    candidates = ['https://test.org/category/this', 'https://www.example.org/1', 'https://otherdomain.org/']
    assert my_urls.filter_unknown_urls(candidates) == ['https://test.org/category/this', 'https://otherdomain.org/']
    # visited or not
    assert url_tuples[0].visited is True and url_tuples[1].visited is True and url_tuples[2].visited is False
    assert my_urls.has_been_visited('http://tovisit.com/page') is True
    assert my_urls.urldict['http://tovisit.com'].all_visited is True
    assert my_urls.filter_unvisited_urls(['http://tovisit.com/page']) == []
    assert my_urls.filter_unvisited_urls(['http://tovisit.com/otherpage']) == ['http://tovisit.com/otherpage']
    assert my_urls.has_been_visited('https://www.other.org/1') is False
    assert my_urls.has_been_visited(url1) is True
    assert my_urls.has_been_visited(f'{example_domain}/this') is False
    assert my_urls.has_been_visited(f'{example_domain}/999') is False
    candidates = [url1, f'{example_domain}/this', f'{example_domain}/999']
    assert my_urls.filter_unvisited_urls(candidates) == [example_domain + '/this', example_domain + '/999']
    assert len(my_urls.find_known_urls(example_domain)) == len(my_urls._load_urls(example_domain)) == 10011
    assert len(my_urls.find_unvisited_urls(example_domain)) == 10009
    assert my_urls.unvisited_websites_number() == 4

    # get download URLs
    downloadable_urls = my_urls.get_download_urls(timelimit=0)
    assert len(downloadable_urls) == 2 and downloadable_urls[0] == 'https://www.example.org/1'
    assert (datetime.now() - my_urls.urldict['https://www.example.org'].timestamp).total_seconds() < 0.25
    assert my_urls.urldict['https://www.example.org'].count == 3
    assert my_urls.urldict['https://test.org'].count == 1
    downloadable_urls = my_urls.get_download_urls()  # limit=10
    assert len(downloadable_urls) == 0
    other_store = UrlStore()
    downloadable_urls = other_store.get_download_urls()
    assert downloadable_urls is None and other_store.done is True

    # schedule
    schedule = other_store.establish_download_schedule()
    assert schedule == []
    # store exhaustion
    other_store = UrlStore()
    other_store.add_urls(['http://domain.fi/page1', 'http://domain.fi/page2', 'http://domain.no/0'])
    schedule = other_store.establish_download_schedule()
    assert len(schedule) == 3
    # reaching buffer limit
    schedule = my_urls.establish_download_schedule(max_urls=1, time_limit=1)
    assert len(schedule) == 1 and round(schedule[0][0]) == 1 and schedule[0][1] == 'https://www.example.org/2'
    schedule = my_urls.establish_download_schedule(max_urls=6, time_limit=1)
    assert len(schedule) == 6 and round(max(s[0] for s in schedule)) == 4
    assert my_urls.urldict['https://www.example.org'].count == 7
    assert my_urls.urldict['https://test.org'].count == 4 == sum(u.visited is True for u in my_urls.urldict['https://test.org'].tuples)
    assert my_urls.download_threshold_reached(8) is False
    assert my_urls.download_threshold_reached(7) is True


def test_dbdump(capsys):
    'Test cases where the URLs are dumped or printed.'

    # database dump
    first_one = UrlStore()
    first_one.add_urls(['http://example.org/print', 'http://print.org/print'])
    assert len(first_one.dump_urls()) == 2

    # database print out
    other_one = UrlStore()
    other_one.add_urls(['http://print.org/print'])
    other_one.print_urls()
    captured = capsys.readouterr()
    assert captured.out.strip() == 'http://print.org/print\tFalse'
    interrupted_one = UrlStore()
    interrupted_one.add_urls(['https://www.example.org/1', 'https://www.example.org/10'])

    # don't test it on Windows
    if os.name != 'nt':
        pid = os.getpid()
        with pytest.raises(SystemExit):
            os.kill(pid, signal.SIGINT)
        captured = capsys.readouterr()
        assert captured.err.strip().endswith('https://www.example.org/10')




"""
Core functions needed to make the module work.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

#import locale
import logging
import re
import sys

#from functools import cmp_to_key
from random import sample


from .clean import normalize_url, scrub_url
from .filters import basic_filter, extension_filter, lang_filter, \
                     path_filter, spam_filter, type_filter, validate_url
from .network import redirection_test
from .settings import BLACKLIST
from .urlstore import UrlStore
from .urlutils import extract_domain, fix_relative_urls, is_external, is_known_link


LOGGER = logging.getLogger(__name__)

FIND_LINKS_REGEX = re.compile(r'<a [^<>]+?>', re.I)
HREFLANG_REGEX = re.compile(r'hreflang=["\']?([a-z-]+)', re.I)
LINK_REGEX = re.compile(r'href=["\']?([^ ]+?)(["\']|[ >])', re.I)


def check_url(url, strict=False, with_redirects=False, language=None, with_nav=False):
    """ Check links for appropriateness and sanity
    Args:
        url: url to check
        strict: set to True for stricter filtering
        with_redirects: set to True for redirection test (per HTTP HEAD request)
        language: set target language (ISO 639-1 codes)
        with_nav: set to True to include navigation pages instead of discarding them

    Returns:
        A tuple consisting of canonical URL and extracted domain

    Raises:
        ValueError, handled in exception.
    """

    # first sanity check
    # use standard parsing library, validate and strip fragments, then normalize
    try:
        # length test
        if basic_filter(url) is False:
            LOGGER.debug('rejected, basic filter: %s', url)
            raise ValueError

        # clean
        url = scrub_url(url)

        # get potential redirect
        if with_redirects is True:
            url = redirection_test(url)
            if url is None:
                LOGGER.debug('rejected, redirection: %s', url)
                raise ValueError

        # spam
        if strict is True and spam_filter(url) is False:
            LOGGER.debug('rejected, spam filter: %s', url)
            raise ValueError
        # structural elements
        if type_filter(url, strict=strict, with_nav=with_nav) is False:
            LOGGER.debug('rejected, type filter: %s', url)
            raise ValueError

        # internationalization and language heuristics in URL
        if (
            language is not None
            and lang_filter(url, language, strict) is False
        ):
            LOGGER.debug('rejected, lang filter: %s', url)
            raise ValueError

        # split and validate
        validation_test, parsed_url = validate_url(url)
        if validation_test is False:
            LOGGER.debug('rejected, validation test: %s', url)
            raise ValueError

        # content filter based on extensions
        if extension_filter(parsed_url.path) is False:
            LOGGER.debug('rejected, extension filter: %s', url)
            raise ValueError

        # strict content filtering
        if strict is True and path_filter(parsed_url.path, parsed_url.query) is False:
            LOGGER.debug('rejected, path filter: %s', url)
            raise ValueError

        # normalize
        url = normalize_url(parsed_url, strict, language)

        # domain info: use blacklist in strict mode only
        if strict is True:
            domain = extract_domain(url, blacklist=BLACKLIST)
        else:
            domain = extract_domain(url)
        if domain is None:
            LOGGER.debug('rejected, domain name: %s', url)
            return None

    # handle exceptions
    except (AttributeError, ValueError):
        LOGGER.debug('discarded URL: %s', url)
        return None

    return url, domain


def sample_urls(input_urls, samplesize, exclude_min=None, exclude_max=None, strict=False, verbose=False):
    '''Sample a list of URLs by domain name, optionally using constraints on their number'''
    # logging
    if verbose is True:
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    else:
        logging.basicConfig(stream=sys.stdout, level=logging.ERROR)
    # deduplicate
    input_urls = list(dict.fromkeys(input_urls))
    # validate
    input_urls = [u for u in input_urls if check_url(u, strict=strict, with_redirects=False) is not None]
    # store
    output_urls = []
    urlstore = UrlStore(compressed=False, language=None, strict=strict)
    urlstore.add_urls(input_urls)
    # iterate
    for domain in urlstore.urldict:  # key=cmp_to_key(locale.strcoll)
        urlpaths = [p.urlpath for p in urlstore._load_urls(domain)]
        # too few or too many URLs
        if (exclude_min is not None and len(urlpaths) < exclude_min or
            exclude_max is not None and len(urlpaths) > exclude_max):
            LOGGER.info('discarded (size): %s\t\turls: %s', domain, len(urlpaths))
            continue
        # copy all URLs
        if len(urlpaths) <= samplesize:
            output_urls.extend([domain + p for p in urlpaths])
            LOGGER.info('%s\t\turls: %s', domain, len(urlpaths))
            continue
        # sample
        mysample = sorted(sample(urlpaths, k=samplesize))
        output_urls.extend([domain + p for p in mysample])
        LOGGER.info('%s\t\turls: %s\tprop.: %s', domain, len(mysample), len(mysample)/len(urlpaths))
    # return gathered URLs
    return output_urls


def extract_links(pagecontent, base_url, external_bool, language=None,
                  strict=True, with_nav=False, redirects=False, reference=None):
    """ Filter links in a HTML document using a series of heuristics
    Args:
        pagecontent: whole page in binary format
        base_url: beginning of the URL, without path, fragment and query
        external: set to True for external links only, False for
                  internal links only
        language: set target language (ISO 639-1 codes)
        strict: set to True for stricter filtering
        with_nav: set to True to include navigation pages instead of discarding them
        with_redirects: set to True for redirection test (per HTTP HEAD request)
        reference: provide a host reference for external/internal evaluation

    Returns:
        A set containing filtered HTTP links checked for sanity and consistency.

    Raises:
        Nothing.
    """
    candidates, validlinks = set(), set()
    if pagecontent is None or pagecontent == '':
        return validlinks
    # define host reference
    if reference is None:
        reference = base_url
    # extract links
    for link in FIND_LINKS_REGEX.findall(pagecontent):
        # https://en.wikipedia.org/wiki/Hreflang
        if language is not None and 'hreflang' in link:
            langmatch = HREFLANG_REGEX.search(link)
            if langmatch and (
                langmatch.group(1).startswith(language) or
                langmatch.group(1) == 'x-default'
                ):
                mymatch = LINK_REGEX.search(link)
                if mymatch:
                    candidates.add(mymatch.group(1))
        # default
        else:
            mymatch = LINK_REGEX.search(link)
            if mymatch:
                candidates.add(mymatch.group(1))
    # filter candidates
    for link in candidates:
        # repair using base
        if not link.startswith('http'):
            link = fix_relative_urls(base_url, link)
        # check
        checked = check_url(link, strict=strict, with_nav=with_nav,
                            with_redirects=redirects, language=language)
        if checked is None:
            continue
        link = checked[0]
        # external/internal links
        if external_bool != is_external(link, reference):
            continue
        if is_known_link(link, validlinks) is True:
            continue
        validlinks.add(link)
    # return
    LOGGER.info('%s links found – %s valid links', len(candidates), len(validlinks))
    return validlinks



"""
Functions related to URL manipulation and extraction of URL parts.
"""

import re

from functools import lru_cache
from urllib.parse import urlparse, ParseResult

from tld import get_fld, get_tld


DOMAIN_REGEX = re.compile(r'^www[0-9]*\.')

@lru_cache(maxsize=1024)
def get_tldinfo(url):
    '''Cached function to extract top-level domain info'''
    return get_tld(url, as_object=True, fail_silently=True)


def extract_domain(url, blacklist=None):
    '''Extract domain name information using top-level domain info'''
    if blacklist is None:
        blacklist = set()
    # new code: Python >= 3.6 with tld module
    tldinfo = get_tldinfo(url)
    # invalid input OR domain TLD blacklist
    if tldinfo is None or tldinfo.domain in blacklist:
        return None
    # return domain
    # this step seems necessary to standardize output
    return DOMAIN_REGEX.sub('', tldinfo.fld)


def _parse(url):
    'Parse a string or use urllib.parse object directly.'
    if isinstance(url, str):
        parsed_url = urlparse(url)
    elif isinstance(url, ParseResult):
        parsed_url = url
    else:
        raise TypeError('wrong input type:', type(url))
    return parsed_url


def get_base_url(url):
    '''Strip URL of some of its parts to get base URL.
       Accepts strings and urllib.parse ParseResult objects.'''
    parsed_url = _parse(url)
    return parsed_url._replace(path='', params='', query='', fragment='').geturl()


def get_host_and_path(url):
    '''Decompose URL in two parts: protocol + host/domain and path.
       Accepts strings and urllib.parse ParseResult objects.'''
    parsed_url = _parse(url)
    host = parsed_url._replace(path='', params='', query='', fragment='')
    path = parsed_url._replace(scheme='', netloc='')
    hostval, pathval = host.geturl(), path.geturl()
    # correction for root/homepage
    if pathval == '':
        pathval = '/'
    if not hostval or not pathval:
        raise ValueError('incomplete URL: %s', url)
    return hostval, pathval


def get_hostinfo(url):
    'Extract domain and host info (protocol + host/domain) from a URL.'
    domainname = extract_domain(url)
    parsed_url = urlparse(url)
    host = parsed_url._replace(path='', params='', query='', fragment='')
    return domainname, host.geturl()


def fix_relative_urls(baseurl, url):
    'Prepend protocol and host information to relative links.'
    if url.startswith('//'):
        return 'https:' + url if baseurl.startswith('https') else 'http:' + url
    if url.startswith('/'):
        # imperfect path handling
        return baseurl + url
    if url.startswith('.'):
        # don't try to correct these URLs
        return baseurl + '/' + re.sub(r'(.+/)+', '', url)
    if not url.startswith('http') and not url.startswith('{'):
        return baseurl + '/' + url
    # todo: handle here
    #if url.startswith('{'):
    return url


def is_external(url, reference, ignore_suffix=True):
    '''Determine if a link leads to another host, takes a reference URL or
       tld/tldextract object as input, returns a boolean'''
    # new code: Python >= 3.6 with tld module
    if ignore_suffix is True:
        try:
            ref_domain, domain = get_tldinfo(reference).domain, \
                                 get_tldinfo(url).domain
        # invalid input
        except AttributeError:
            return True
    else:
        ref_domain, domain = get_fld(reference, fail_silently=True), get_fld(url, fail_silently=True)
    # comparison
    return domain != ref_domain


def is_known_link(link, known_links):
    "Compare the link and its possible variants to the existing URL base."
    # easy check
    if link in known_links:
        return True
    # trailing slash
    test1 = link.rstrip('/')
    test2 = test1 + '/'
    if test1 in known_links or test2 in known_links:
        return True
    # http/https + trailing slash
    if link[:4] == 'http':
        if link[:5] == 'https':
            testlink = link[:4] + link[5:]
        else:
            testlink = ''.join([link[:4], 's', link[4:]])
        test1, test2 = testlink.rstrip('/'), testlink.rstrip('/') + '/'
        return testlink in known_links or test1 in known_links or test2 in known_links
    return False



"""
Bundles functions needed to target text content and validate the input.
"""


## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import logging
import re

from urllib.parse import urlparse

from langcodes import Language, tag_is_valid

from .langinfo import COUNTRY_CODES, LANGUAGE_CODES


LOGGER = logging.getLogger(__name__)

# content filters
WORDPRESS_CONTENT_FILTER = re.compile(r'/(?:page|seite|user|search|gallery|gall?erie|labels|archives|uploads|modules|attachment)/|/(?:tags?|schlagwort|category|cat|kategorie|kat|auth?or)/[^/]+/?$', re.IGNORECASE)
PARAM_FILTER = re.compile(r'\.(atom|json|css|xml|js|jpg|jpeg|png|gif|tiff|pdf|ogg|mp3|m4a|aac|avi|mp4|mov|webm|flv|ico|pls|zip|tar|gz|iso|swf)\b', re.IGNORECASE)   # (?=[&?])
ADULT_FILTER = re.compile(r'\b(?:adult|amateur|arsch|cams?|cash|fick|gangbang|incest|porn|sexyeroti[ck]|sexcam|swinger|xxx|bild\-?kontakte)\b', re.IGNORECASE)  # ass|orgasm ?
UNSUITABLE_FILTER = re.compile(r'\b(?:add?s?|banner|doubleclick|livestream|tradedoubler)\b|/oembed\b')
VIDEOS_FILTER = re.compile(r'\b(?:live|videos?)\b', re.IGNORECASE)

# language filter
PATH_LANG_FILTER = re.compile(r'(?:https?://[^/]+/)([a-z]{2})([_-][a-z]{2,3})?(?:/)', re.IGNORECASE)
ALL_PATH_LANGS = re.compile(r'(?:/)([a-z]{2})([_-][a-z]{2})?(?:/)', re.IGNORECASE)
HOST_LANG_FILTER = re.compile(r'https?://([a-z]{2})\.(?:[^.]+)\.(?:[^.]+)/', re.IGNORECASE)

# navigation/crawls
NAVIGATION_FILTER = re.compile(r'/(archives|auth?or|cat|category|kat|kategorie|page|schlagwort|seite|tags?|topics?|user)/|\?p=[0-9]+', re.IGNORECASE)
NOTCRAWLABLE = re.compile(r'/(login|impressum|imprint)(\.[a-z]{3,4})?/?$|/login\?|/(javascript:|mailto:|tel\.?:|whatsapp:)', re.IGNORECASE)
# |/(www\.)?(facebook\.com|google\.com|instagram\.com|twitter\.com)/
INDEX_PAGE_FILTER = re.compile(r'.{0,5}/index(\.[a-z]{3,4})?/?$', re.IGNORECASE)

# document types
EXTENSION_REGEX = re.compile(r'\.[a-z]{2,5}$')
# https://en.wikipedia.org/wiki/List_of_file_formats#Web_page
WHITELISTED_EXTENSIONS = ('.adp', '.amp', '.asp', '.aspx', '.cfm', '.cgi', '.do', '.htm', 'html', '.htx', '.jsp', '.mht', '.mhtml', '.php', '.php3', '.php4', '.php5', '.phtml', '.pl', '.shtml', '.stm', '.txt', '.xhtml', '.xml')

# territories whitelist
# see also: https://babel.pocoo.org/en/latest/api/languages.html
# get_official_languages('ch')
LANGUAGE_MAPPINGS = {
    'de': {'at', 'ch', 'de', 'li'},  # 'be', 'it'
    'en': {'au', 'ca', 'en', 'gb', 'ie', 'nz', 'us'},
    'fr': {'be', 'ca', 'ch', 'fr', 'tn'},  # , 'lu', ...
}


def basic_filter(url):
    '''Filter URLs based on basic formal characteristics'''
    return bool(url.startswith('http') and 10 <= len(url) < 500)


def extension_filter(urlpath):
    '''Filter based on file extension'''
    return bool(
        not EXTENSION_REGEX.search(urlpath)
        or urlpath.endswith(WHITELISTED_EXTENSIONS)
    )


def langcodes_score(language, segment, score):
    '''Use langcodes on selected URL segments and integrate
       them into a score.'''
    # see also: https://babel.pocoo.org/en/latest/locale.html
    # test if the code looks like a country or a language
    if segment[:2] not in COUNTRY_CODES and segment[:2] not in LANGUAGE_CODES:
        return score
    # test if tag is valid (caution: private codes are)
    if tag_is_valid(segment):
        # try to identify language code
        identified = Language.get(segment).language
        # see if it matches
        if identified is not None:
            LOGGER.debug('langcode %s found in URL segment %s', identified, segment)
            if identified != language:
                score -= 1
            else:
                score += 1
    return score


def lang_filter(url, language=None, strict=False):
    '''Heuristics targeting internationalization and linguistic elements.
       Based on a score.'''
    # sanity check
    if language is None:
        return True
    # init score
    score = 0
    # first test: internationalization in URL path
    match = PATH_LANG_FILTER.match(url)
    if match:
        # look for other occurrences
        occurrences = ALL_PATH_LANGS.findall(url)
        if len(occurrences) == 1:
            score = langcodes_score(language, match.group(1), score)
        elif len(occurrences) == 2:
            for occurrence in occurrences:
                score = langcodes_score(language, occurrence, score)
        # don't perform the test if there are too many candidates: > 2
    # second test: prepended language cues
    if strict is True and language in LANGUAGE_MAPPINGS:
        match = HOST_LANG_FILTER.match(url)
        if match:
            candidate = match.group(1).lower()
            LOGGER.debug('candidate lang %s found in URL', candidate)
            if candidate in LANGUAGE_MAPPINGS[language]:
                score += 1
            else:
                score -= 1
    # determine test result
    return score >= 0


def path_filter(urlpath, query):
    '''Filters based on URL path: index page, imprint, etc.'''
    if NOTCRAWLABLE.search(urlpath):
        return False
    if INDEX_PAGE_FILTER.match(urlpath) and len(query) == 0:
        #print('#', urlpath, INDEX_PAGE_FILTER.match(urlpath), query)
        return False
    return True


def spam_filter(url):
    '''Try to filter out spam and adult websites'''
    # TODO: to improve!
    #for exp in (''):
    #    if exp in url:
    #        return False
    return not ADULT_FILTER.search(url)


def type_filter(url, strict=False, with_nav=False):
    '''Make sure the target URL is from a suitable type (HTML page with primarily text)'''
    try:
        # feeds
        if url.endswith(('/feed', '/rss')):
            raise ValueError
        # wordpress structure
        if WORDPRESS_CONTENT_FILTER.search(url) and (
            with_nav is not True or not is_navigation_page(url)
        ):
            raise ValueError
        # hidden in parameters
        if strict is True and PARAM_FILTER.search(url):
            raise ValueError
        # not suitable, ads and embedded content
        if UNSUITABLE_FILTER.search(url):
            raise ValueError
        if strict is True and VIDEOS_FILTER.search(url):
            raise ValueError
    except ValueError:
        return False
    # default
    return True


def validate_url(url):
    '''Parse and validate the input'''
    try:
        parsed_url = urlparse(url)
    except ValueError:
        return False, None
    if not bool(parsed_url.scheme) or parsed_url.scheme not in (
        'http',
        'https',
    ):
        return False, None
    if len(parsed_url.netloc) < 5 or \
       (parsed_url.netloc.startswith('www.') and len(parsed_url.netloc) < 8):
        return False, None
    # default
    return True, parsed_url


def is_navigation_page(url):
    '''Determine if the URL is related to navigation and overview pages
       rather than content pages, e.g. /page/1 vs. article page.'''
    return bool(NAVIGATION_FILTER.search(url))


def is_not_crawlable(url):
    '''Run tests to check if the URL may lead to deep web or pages
       generally not usable in a crawling context.'''
    return bool(NOTCRAWLABLE.search(url))



"""
Defines a URL store which holds URLs along with relevant information.
"""

import bz2
import logging
import pickle
import signal
import sys

from collections import defaultdict, deque
from datetime import datetime, timedelta
from threading import Lock

from .filters import lang_filter, validate_url
from .urlutils import get_host_and_path, is_known_link


LOGGER = logging.getLogger(__name__)


class DomainEntry:
    "Class to record host-related information and URL paths."
    __slots__ = ('all_visited', 'count', 'rules', 'timestamp', 'tuples')

    def __init__(self):
        self.all_visited = False
        self.count = 0
        self.rules = None
        self.tuples = deque()
        self.timestamp = None


class UrlPathTuple:
    "Class storing information for URL paths relative to a domain/host."
    __slots__ = ('urlpath', 'visited')

    def __init__(self, urlpath, visited):
        self.urlpath = urlpath
        self.visited = visited


class UrlStore:
    "Defines a class to store domain-classified URLs and perform checks against it."
    __slots__ = ('compressed', 'done', 'language', 'strict', 'urldict', '_lock')

    def __init__(self, compressed=False, language=None, strict=False):
        self.compressed = compressed
        self.done = False
        self.language = language
        self.strict = strict
        self.urldict = defaultdict(DomainEntry)
        self._lock = Lock()

        def dump_unvisited_urls(num, frame):
            LOGGER.warning('Processing interrupted, dumping unvisited URLs from %s hosts', len(self.urldict))
            for domain in self.urldict:
                print('\n'.join([domain + u.urlpath for u in self._load_urls(domain) if u.visited is False]), file=sys.stderr)
            sys.exit()

        # don't use the following on Windows
        if not sys.platform.startswith('win'):
            signal.signal(signal.SIGINT, dump_unvisited_urls)
            signal.signal(signal.SIGTERM, dump_unvisited_urls)

    #def _filter_urlpaths(self, domain, urls):
    #    if self.validation is True or self.language is not None:
    #        return [u for u in urls if self._filter_url(domain + u) is True]
    #    return urls

    def _buffer_urls(self, data, visited=False):
        inputdict = defaultdict(deque)
        for url in list(dict.fromkeys(data)):
            # segment URL and add to domain dictionary
            try:
                # validate
                validation_result, parsed_url = validate_url(url)
                if validation_result is False:
                    LOGGER.debug('Invalid URL: %s', url)
                    raise ValueError
                # filter
                if self.language is not None and lang_filter(url, self.language, self.strict) is False:
                    LOGGER.debug('Wrong language: %s', url)
                    raise ValueError
                hostinfo, urlpath = get_host_and_path(parsed_url)
                inputdict[hostinfo].append(UrlPathTuple(urlpath, visited))
            except (TypeError, ValueError):
                LOGGER.warning('Discarding URL: %s', url)
        return inputdict

    def _load_urls(self, domain):
        value = self.urldict[domain].tuples
        if isinstance(value, bytes):
            return pickle.loads(bz2.decompress(value))
        return value

    def _store_urls(self, domain, to_right=None, timestamp=None, to_left=None):
        if domain in self.urldict:
            urls = self._load_urls(domain)
            known = {u.urlpath for u in urls}
        else:
            urls = deque()
            known = set()
        # check if the link or its variants are known
        if to_right is not None:
            urls.extend(t for t in to_right if not is_known_link(t.urlpath, known))
        if to_left is not None:
            urls.extendleft(t for t in to_left if not is_known_link(t.urlpath, known))
        # use lock
        with self._lock:
            # compression
            if self.compressed is True:
                self.urldict[domain].tuples = bz2.compress(pickle.dumps(urls, protocol=4))
            else:
                self.urldict[domain].tuples = urls
            # adjust all_visited status
            self.urldict[domain].all_visited = all(u.visited is True for u in urls)
            # timestamp/backoff value
            if timestamp is not None:
                self.urldict[domain].timestamp = timestamp

    def _search_urls(self, urls, switch=None):
        # init
        last_domain, known_paths = None, set()
        remaining_urls = {u: None for u in urls}
        # iterate
        for url in sorted(remaining_urls):
            hostinfo, urlpath = get_host_and_path(url)
            # examine domain
            if hostinfo != last_domain:
                last_domain = hostinfo
                if switch == 1:
                    known_paths = {u.urlpath for u in self._load_urls(hostinfo)}
                elif switch == 2:
                    known_paths = {u.urlpath: u.visited for u in self._load_urls(hostinfo)}
            # run checks: case 1: the path matches, case 2: visited URL
            if urlpath in known_paths and (
                switch == 1 or (switch == 2 and known_paths[urlpath] is True)
            ):
                del remaining_urls[url]
        # preserve input order
        return list(remaining_urls)

    def _timestamp(self, domain):
        return self.urldict[domain].timestamp

    # URL MANIPULATION AND INFO

    def add_urls(self, urls=None, appendleft=None, visited=False):
        """Add a list of URLs to the (possibly) existing one.
        Optional: append certain URLs to the left,
        specify if the URLs have already been visited."""
        if urls:
            for host, urltuples in self._buffer_urls(urls, visited).items():
                self._store_urls(host, to_right=urltuples)
        if appendleft:
            for host, urltuples in self._buffer_urls(appendleft, visited).items():
                self._store_urls(host, to_left=urltuples)

    def is_known(self, url):
        "Check if the given URL has already been stored."
        hostinfo, urlpath = get_host_and_path(url)
        values = self._load_urls(hostinfo)
        # returns False if domain or URL is new
        return urlpath in {u.urlpath for u in values}

    def find_known_urls(self, domain):
        """Get all already known URLs for the given domain (ex. "https://example.org")."""
        values = self._load_urls(domain)
        return [domain + u.urlpath for u in values]

    def filter_unknown_urls(self, urls):
        "Take a list of URLs and return the currently unknown ones."
        return self._search_urls(urls, switch=1)

    def get_known_domains(self):
        "Return all known domains as a list."
        return list(self.urldict)

    def is_exhausted_domain(self, domain):
        "Tell if all known URLs for the website have been visited."
        if domain in self.urldict:
            return self.urldict[domain].all_visited
        raise KeyError('website not in store')

    # URL-BASED QUERIES

    def has_been_visited(self, url):
        "Check if the given URL has already been visited.."
        hostinfo, urlpath = get_host_and_path(url)
        values = self._load_urls(hostinfo)
        known_urlpaths = {u.urlpath: u.visited for u in values}
        # defaults to None, thus False
        return known_urlpaths.get(urlpath) or False

    def find_unvisited_urls(self, domain):
        "Get all unvisited URLs for the given domain."
        values = self._load_urls(domain)
        return [domain + u.urlpath for u in values if u.visited is False]

    def filter_unvisited_urls(self, urls):
        "Take a list of URLs and return the currently unvisited ones."
        return self._search_urls(urls, switch=2)

    def unvisited_websites_number(self):
        "Return the number of websites for which there are still URLs to visit."
        return len([d for d in self.urldict if self.urldict[d].all_visited is False])

    # DOWNLOADS

    def get_url(self, domain):
        "Retrieve a single URL and consider it to be visited (with corresponding timestamp)."
        # not fully used
        if self.urldict[domain].all_visited is False:
            url_tuples = self._load_urls(domain)
            # get first non-seen url
            for url in url_tuples:
                if url.visited is False:
                    url.visited = True
                    with self._lock:
                        self.urldict[domain].count += 1
                    self._store_urls(domain, url_tuples, timestamp=datetime.now())
                    return domain + url.urlpath
        # nothing to draw from
        with self._lock:
            self.urldict[domain].all_visited = True
        return None

    def get_download_urls(self, timelimit=10):
        """Get a list of immediately downloadable URLs according to the given
           time limit per domain."""
        with self._lock:
            potential = [d for d in self.urldict if self.urldict[d].all_visited is False]
        if not potential:
            self.done = True
            return None
        targets = []
        for domain in potential:
            timestamp = self._timestamp(domain)
            if timestamp is None or (datetime.now() - timestamp).total_seconds() > timelimit:
                targets.append(domain)
        # get corresponding URLs and filter out None values
        return list(filter(None, [self.get_url(domain) for domain in targets]))

    def establish_download_schedule(self, max_urls=100, time_limit=10):
        """Get up to the specified number of URLs along with a suitable
           backoff schedule (in seconds)."""
        # see which domains are free
        with self._lock:
            potential = [d for d in self.urldict if self.urldict[d].all_visited is False]
            if not potential:
                self.done = True
                return []
        # variables init
        per_domain = max_urls // len(potential) or 1
        targets = []
        # iterate potential domains
        for domain in potential:
            url_tuples = self._load_urls(domain)
            # load urls
            urlpaths = []
            # get first non-seen urls
            for url in url_tuples:
                if len(urlpaths) >= per_domain or (len(targets) + len(urlpaths)) >= max_urls:
                    break
                if url.visited is False:
                    urlpaths.append(url.urlpath)
                    url.visited = True
                    with self._lock:
                        self.urldict[domain].count += 1
            # determine timestamps
            now = datetime.now()
            original_timestamp = self._timestamp(domain)
            if original_timestamp is None or (now - original_timestamp).total_seconds() > time_limit:
                schedule_secs = 0
            else:
                schedule_secs = time_limit - float(f'{(now - original_timestamp).total_seconds():.2f}')
            for urlpath in urlpaths:
                targets.append((schedule_secs, domain + urlpath))
                schedule_secs += time_limit
            # calculate difference and offset last addition
            total_diff = now + timedelta(0, schedule_secs - time_limit)
            # store new info
            self._store_urls(domain, url_tuples, timestamp=total_diff)
        # sort by first tuple element (time in secs)
        return sorted(targets, key=lambda x: x[0])

    def get_rules(self, website):
        "Return the stored crawling rules for the given website."
        return self.urldict[website].rules

    # GENERAL INFO

    def total_url_number(self):
        "Find number of all URLs in store."
        return sum(len(self.urldict[d].tuples) for d in self.urldict)

    def download_threshold_reached(self, threshold):
        "Find out if the download limit (in seconds) has been reached for one of the websites in store."
        return any(self.urldict[d].count >= threshold for d in self.urldict)

    def dump_urls(self):
        "Return a list of all known URLs."
        urls = []
        for domain in self.get_known_domains():
            urls.extend(self.find_known_urls(domain))
        return urls

    def print_urls(self):
        "Print all URLs in store (URL + TAB + visited or not)."
        for domain in self.get_known_domains():
            print('\n'.join([domain + u.urlpath + '\t' + str(u.visited) for u in self._load_urls(domain)]))



"""
URL filter and manipulation tools
https://github.com/adbar/courlan
"""

import re

from pathlib import Path
from setuptools import setup



def get_version(package):
    "Return package version as listed in `__version__` in `init.py`"
    # version = Path(package, '__init__.py').read_text() # Python >= 3.5
    with open(str(Path(package, '__init__.py')), 'r', encoding='utf-8') as filehandle:
        initfile = filehandle.read()
    return re.search('__version__ = [\'"]([^\'"]+)[\'"]', initfile).group(1)


def get_long_description():
    "Return the README"
    with open('README.rst', 'r', encoding='utf-8') as filehandle:
        long_description = filehandle.read()
    #long_description += "\n\n"
    #with open("CHANGELOG.md", encoding="utf8") as f:
    #    long_description += f.read()
    return long_description


setup(
    name='courlan',
    version=get_version('courlan'),
    description='Clean, filter and sample URLs to optimize data collection – includes spam, content type and language filters.',
    long_description=get_long_description(),
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        #'Development Status :: 6 - Mature',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: Linguistic',
    ],
    keywords=['cleaner', 'crawler', 'preprocessing', 'url-parsing', 'url-manipulation', 'urls', 'validation', 'webcrawling'],
    url='https://github.com/adbar/courlan',
    author='Adrien Barbaresi',
    author_email='barbaresi@bbaw.de',
    license='GPLv3+',
    packages=['courlan'],
    project_urls={
        "Blog": "https://adrien.barbaresi.eu/blog/",  # /tag/courlan.html
        "Tracker": "https://github.com/adbar/courlan/issues",
    },
    #package_data={},
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'langcodes >= 3.3.0',
        'tld >= 0.12.6',
        'urllib3 >= 1.26, < 2',
    ],
    #extras_require=extras,
    entry_points = {
        'console_scripts': ['courlan=courlan.cli:main'],
    },
    # platforms='any',
    tests_require=['pytest'],
    zip_safe=False,
)



"""
Unit tests for the courlan package.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import io
import logging
import os
import subprocess
import sys
import tempfile

from contextlib import redirect_stdout
from unittest.mock import patch
from urllib.parse import ParseResult

import pytest

from courlan import cli
from courlan import clean_url, normalize_url, scrub_url, check_url, is_external, sample_urls, validate_url, extract_links, extract_domain, fix_relative_urls, get_base_url, get_host_and_path, get_hostinfo, is_navigation_page, is_not_crawlable, lang_filter
from courlan.filters import extension_filter, path_filter, spam_filter, type_filter
from courlan.urlutils import _parse, is_known_link


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
RESOURCES_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')



def test_baseurls():
    assert get_base_url('https://example.org/') == 'https://example.org'
    assert get_base_url('https://example.org/test.html?q=test#frag') == 'https://example.org'
    assert get_base_url('example.org') == ''


def test_fix_relative():
    assert fix_relative_urls('https://example.org', 'page.html') == 'https://example.org/page.html'
    assert fix_relative_urls('http://example.org', '//example.org/page.html') == 'http://example.org/page.html'
    assert fix_relative_urls('https://example.org', './page.html') == 'https://example.org/page.html'
    assert fix_relative_urls('https://example.org', '/page.html') == 'https://example.org/page.html'
    # fixing partial URLs
    assert fix_relative_urls('https://example.org', 'https://example.org/test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('https://example.org', '/test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('https://example.org', '//example.org/test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('http://example.org', '//example.org/test.html') == 'http://example.org/test.html'
    assert fix_relative_urls('https://example.org', 'test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('https://example.org', '../../test.html') == 'https://example.org/test.html'


def test_scrub():
    # clean: scrub + normalize
    assert clean_url(5) is None
    assert clean_url('ø\xaa') == 'øª'
    # scrub
    assert scrub_url('  https://www.dwds.de') == 'https://www.dwds.de'
    assert scrub_url('<![CDATA[https://www.dwds.de]]>') == 'https://www.dwds.de'
    assert scrub_url('https://www.dwds.de/test?param=test&amp;other=test') == 'https://www.dwds.de/test?param=test&other=test'
    assert scrub_url('https://www.dwds.de/garbledhttps://www.dwds.de/') == 'https://www.dwds.de/garbled'
    assert scrub_url('https://g__https://www.dwds.de/') == 'https://www.dwds.de'
    # exception for archive URLs
    assert scrub_url('https://web.archive.org/web/20131021165347/https://www.imdb.com/') == 'https://web.archive.org/web/20131021165347/https://www.imdb.com'
    # social sharing
    assert scrub_url('https://twitter.com/share?&text=Le%20sabre%20de%20bambou%20%232&via=NouvellesJapon&url=https://nouvellesdujapon.com/le-sabre-de-bambou-2') == 'https://nouvellesdujapon.com/le-sabre-de-bambou-2'
    assert scrub_url('https://www.facebook.com/sharer.php?u=https://nouvellesdujapon.com/le-sabre-de-bambou-2') == 'https://nouvellesdujapon.com/le-sabre-de-bambou-2'
    # end of URL
    assert scrub_url('https://www.test.com/&') == 'https://www.test.com'
    # white space
    assert scrub_url('\x19https://www.test.com/\x06') == 'https://www.test.com'
    # markup
    assert scrub_url('https://www.test.com/</a>') == 'https://www.test.com'
    # garbled URLs e.g. due to quotes
    assert scrub_url('https://www.test.com/"' + '<p></p>'*100) == 'https://www.test.com'
    assert scrub_url('https://www.test.com/"' * 50) != 'https://www.test.com'
    # simply too long, left untouched
    my_url = 'https://www.test.com/' + 'abcdefg' * 100
    assert scrub_url(my_url) == my_url


def test_extension_filter():
    validation_test, parsed_url = validate_url('http://www.example.org/test.js')
    assert extension_filter(parsed_url.path) is False
    validation_test, parsed_url = validate_url('http://goodbasic.com/GirlInfo.aspx?Pseudo=MilfJanett')
    assert extension_filter(parsed_url.path) is True
    validation_test, parsed_url = validate_url('https://www.familienrecht-allgaeu.de/de/vermoegensrecht.amp')
    assert extension_filter(parsed_url.path) is True
    validation_test, parsed_url = validate_url('http://www.example.org/test.shtml')
    assert extension_filter(parsed_url.path) is True
    validation_test, parsed_url = validate_url('http://de.artsdot.com/ADC/Art.nsf/O/8EWETN')
    assert extension_filter(parsed_url.path) is True
    validation_test, parsed_url = validate_url('http://de.artsdot.com/ADC/Art.nsf?param1=test')
    assert extension_filter(parsed_url.path) is False
    validation_test, parsed_url = validate_url('http://www.example.org/test.xhtml?param1=this')
    assert extension_filter(parsed_url.path) is True
    validation_test, parsed_url = validate_url('http://www.example.org/test.php5')
    assert extension_filter(parsed_url.path) is True
    validation_test, parsed_url = validate_url('http://www.example.org/test.php6')
    assert extension_filter(parsed_url.path) is True


def test_spam_filter():
    assert spam_filter('http://www.example.org/cams/test.html') is False
    assert spam_filter('http://www.example.org/test.html') is True


def test_type_filter():
    assert type_filter('http://www.example.org/feed') is False
    # straight category
    assert type_filter('http://www.example.org/category/123') is False
    # post simply filed under a category
    assert type_filter('http://www.example.org/category/tropes/time-travel') is True
    assert type_filter('http://www.example.org/test.xml?param=test', strict=True) is False
    assert type_filter('http://www.example.org/test.asp') is True
    assert type_filter('http://ads.example.org/') is False
    # -video- vs. /video/
    assert type_filter('http://my-videos.com/') is True
    assert type_filter('http://my-videos.com/', strict=True) is False
    assert type_filter('http://example.com/video/1', strict=True) is False
    assert type_filter('http://example.com/new-video-release') is True
    assert type_filter('http://example.com/new-video-release', strict=True) is False
    # tags
    assert type_filter('https://de.thecitizen.de/tag/anonymity/') is False
    assert type_filter('https://de.thecitizen.de/tags/anonymity/') is False
    # author
    assert type_filter('http://www.example.org/author/abcde') is False
    assert type_filter('http://www.example.org/autor/abcde/') is False
    # misc
    assert type_filter('http://www.bmbwk.gv.at/forschung/fps/gsk/befragung.xml?style=text') is True
    #assert type_filter('http://www.aec.at/de/archives/prix_archive/prix_projekt.asp?iProjectID=11118') is True


def test_path_filter():
    assert check_url('http://www.case-modder.de/index.php?sec=artikel&id=68&page=1', strict=True) is not None
    assert check_url('http://www.case-modder.de/index.php', strict=True) is None


def test_lang_filter():
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', None) is True
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'de') is False
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'fr') is True
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'en') is False
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'es') is False
    assert lang_filter('https://www.sitemaps.org/en_GB/protocol.html', 'en') is True
    assert lang_filter('https://www.sitemaps.org/en_GB/protocol.html', 'de') is False
    assert lang_filter('https://en.wikipedia.org/', 'de', strict=True) is False
    assert lang_filter('https://en.wikipedia.org/', 'de', strict=False) is True
    assert lang_filter('https://de.wikipedia.org/', 'de', strict=True) is True
    assert lang_filter('http://de.musclefood.com/neu/neue-nahrungsergaenzungsmittel.html', 'de', strict=True) is True
    assert lang_filter('http://de.musclefood.com/neu/neue-nahrungsergaenzungsmittel.html', 'fr', strict=True) is False
    assert lang_filter('http://ch.postleitzahl.org/sankt_gallen/liste-T.html', 'fr') is True
    assert lang_filter('http://ch.postleitzahl.org/sankt_gallen/liste-T.html', 'de') is True
    # to complete when language mappings are more extensive
    # assert lang_filter('http://ch.postleitzahl.org/sankt_gallen/liste-T.html', 'es') is False
    # disturbing path sub-elements
    assert lang_filter('http://www.uni-rostock.de/fakult/philfak/fkw/iph/thies/mythos.html', 'de') is True
    assert lang_filter('http://stifter.literature.at/witiko/htm/h15-22b.html', 'de') is True
    assert lang_filter('http://stifter.literature.at/doc/witiko/h15-22b.html', 'de') is True
    assert lang_filter('http://stifter.literature.at/nl/witiko/h15-22b.html', 'de') is False
    assert lang_filter('http://stifter.literature.at/de_DE/witiko/h15-22b.html', 'de') is True
    assert lang_filter('http://stifter.literature.at/en_US/witiko/h15-22b.html', 'de') is False
    assert lang_filter('http://www.stiftung.koerber.de/bg/recherche/de/beitrag.php?id=15132&refer=', 'de') is True
    assert lang_filter('http://www.solingen-internet.de/si-hgw/eiferer.htm', 'de') is True
    assert lang_filter('http://ig.cs.tu-berlin.de/oldstatic/w2000/ir1/aufgabe2/ir1-auf2-gr16.html', 'de', strict=True) is True
    assert lang_filter('http://ig.cs.tu-berlin.de/oldstatic/w2000/ir1/aufgabe2/ir1-auf2-gr16.html', 'de', strict=False) is True
    assert lang_filter('http://bz.berlin1.de/kino/050513/fans.html', 'de', strict=False) is True
    assert lang_filter('http://bz.berlin1.de/kino/050513/fans.html', 'de', strict=True) is False
    # assert lang_filter('http://www.verfassungen.de/ch/basel/verf03.htm'. 'de') is True
    # assert lang_filter('http://www.uni-stuttgart.de/hi/fnz/lehrveranst.html', 'de') is True
    # http://www.wildwechsel.de/ww/front_content.php?idcatart=177&lang=4&client=6&a=view&eintrag=100&a=view&eintrag=0&a=view&eintrag=20&a=view&eintrag=80&a=view&eintrag=20


def test_navigation():
    assert is_navigation_page('https://test.org/') is False
    assert is_navigation_page('https://test.org/page/1') is True
    assert is_navigation_page('https://test.org/?p=11') is True
    assert is_not_crawlable('https://test.org/login') is True
    assert is_not_crawlable('https://test.org/login/') is True
    assert is_not_crawlable('https://test.org/login.php') is True
    assert is_not_crawlable('https://test.org/page') is False


def test_validate():
    assert validate_url('http://www.test[.org/test')[0] is False
    # assert validate_url('http://www.test.org:7ERT/test')[0] is False
    assert validate_url('ntp://www.test.org/test')[0] is False
    assert validate_url('ftps://www.test.org/test')[0] is False
    assert validate_url('http://t.g/test')[0] is False
    assert validate_url('http://test.org/test')[0] is True


def test_normalization():
    assert normalize_url('HTTPS://WWW.DWDS.DE/') == 'https://www.dwds.de/'
    assert normalize_url('http://test.net/foo.html#bar', strict=True) == 'http://test.net/foo.html'
    assert normalize_url('http://test.net/foo.html#bar', strict=False) == 'http://test.net/foo.html#bar'
    assert normalize_url('http://test.net/foo.html#:~:text=night-,vision', strict=True) == 'http://test.net/foo.html'
    assert normalize_url('http://www.example.org:80/test.html') == 'http://www.example.org/test.html'
    assert normalize_url('https://hanxiao.io//404.html') == 'https://hanxiao.io/404.html'


def test_qelems():
    assert normalize_url('http://test.net/foo.html?utm_source=twitter') == 'http://test.net/foo.html?utm_source=twitter'
    assert normalize_url('http://test.net/foo.html?utm_source=twitter', strict=True) == 'http://test.net/foo.html'
    assert normalize_url('http://test.net/foo.html?utm_source=twitter&post=abc&page=2') == 'http://test.net/foo.html?page=2&post=abc&utm_source=twitter'
    assert normalize_url('http://test.net/foo.html?utm_source=twitter&post=abc&page=2', strict=True) == 'http://test.net/foo.html?page=2&post=abc'
    assert normalize_url('http://test.net/foo.html?page=2&itemid=10&lang=en') == 'http://test.net/foo.html?itemid=10&lang=en&page=2'
    with pytest.raises(ValueError):
        assert normalize_url('http://test.net/foo.html?page=2&lang=en', language='de')
        assert normalize_url('http://www.evolanguage.de/index.php?page=deutschkurse_fuer_aerzte&amp;language=ES', language='de')


def test_urlcheck():
    assert check_url('AAA') is None
    assert check_url('1234') is None
    assert check_url('http://ab') is None
    assert check_url('ftps://example.org/') is None
    assert check_url('http://t.g/test') is None
    assert check_url('https://www.dwds.de/test?param=test&amp;other=test', strict=True) == ('https://www.dwds.de/test', 'dwds.de')
    assert check_url('http://example.com/index.html#term', strict=True) is None
    assert check_url('http://example.com/index.html#term', strict=False)[0] == 'http://example.com/index.html#term'
    assert check_url('http://example.com/test.js') is None
    assert check_url('http://twitter.com/', strict=True) is None
    assert check_url('http://twitter.com/', strict=False) is not None
    assert check_url('https://www.httpbin.org/status/200', with_redirects=True) == ('https://www.httpbin.org/status/200', 'httpbin.org')
    #assert check_url('https://www.httpbin.org/status/302', with_redirects=True) == ('https://www.httpbin.org/status/302', 'httpbin.org')
    assert check_url('https://www.httpbin.org/status/404', with_redirects=True) is None
    assert check_url('https://www.ht.or', with_redirects=True) is None
    # recheck type and spam filters
    assert check_url('http://example.org/code/oembed/') is None
    assert check_url('http://cams.com/', strict=False) == ('http://cams.com', 'cams.com')
    assert check_url('http://cams.com/', strict=True) is None
    assert check_url('https://denkiterm.wordpress.com/impressum/', strict=True) is None
    assert check_url('http://www.fischfutter-index.de/improvit-trocken-frostfutter-fur-fast-alle-fische/', strict=True) is not None
    # language and internationalization
    assert check_url('http://example.com/test.html?lang=en', language='de') is None
    assert check_url('http://example.com/test.html?lang=en', language=None) is not None
    assert check_url('http://example.com/test.html?lang=en', language='en') is not None
    assert check_url('http://example.com/de/test.html', language='de') is not None
    assert check_url('http://example.com/en/test.html', language='de') is None
    assert check_url('http://example.com/en/test.html', language=None) is not None
    assert check_url('http://example.com/en/test.html', language='en') is not None
    assert check_url('https://www.myswitzerland.com/de-ch/erlebnisse/veranstaltungen/wild-im-sternen/', language='de') is not None
    assert check_url('https://www.myswitzerland.com/en-id/accommodations/other-types-of-accommodations/on-the-farm/farm-experiences-search/', language='en') is not None
    assert check_url('https://www.myswitzerland.com/EN-ID/accommodations/other-types-of-accommodations/on-the-farm/farm-experiences-search/', language='en') is not None
    # impressum and index
    assert check_url('http://www.example.org/index', strict=True) is None
    assert check_url('http://www.example.org/index.html', strict=True) is None
    assert check_url('http://concordia-hagen.de/impressum.html', strict=True) is None
    assert check_url('http://concordia-hagen.de/de/impressum', strict=True) is None
    assert check_url('http://parkkralle.de/detail/index/sArticle/2704', strict=True) is not None
    assert check_url('https://www.katholisch-in-duisdorf.de/kontakt/links/index.html', strict=True) is not None
    assert check_url('{mylink}') is None
    assert check_url('https://de.nachrichten.yahoo.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html', language='de') is not None
    assert check_url('https://de.nachrichten.yahoo.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html', language='de', strict=True) is None   
    # assert check_url('https://de.nachrichten.yahoo.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html', language='en') is None
    # assert check_url('http://www.immobilienscout24.de/de/ueberuns/presseservice/pressestimmen/2_halbjahr_2000.jsp;jsessionid=287EC625A45BD5A243352DD8C86D25CC.worker2', language='de', strict=True) is not None


def test_urlutils():
    '''Test URL manipulation tools'''
    # domain extraction
    assert extract_domain('https://httpbin.org/') == 'httpbin.org'
    # url parsing
    result = _parse('https://httpbin.org/')
    assert isinstance(result, ParseResult)
    newresult = _parse(result)
    assert isinstance(result, ParseResult)
    with pytest.raises(TypeError):
        result = _parse(float(1.23))

    assert get_base_url('https://example.org/path') == 'https://example.org'
    with pytest.raises(ValueError):
        assert get_host_and_path('123') is None
    assert get_host_and_path('https://example.org/path') == ('https://example.org', '/path')
    assert get_host_and_path('https://example.org/') == ('https://example.org', '/')
    assert get_host_and_path('https://example.org') == ('https://example.org', '/')
    assert get_hostinfo('https://httpbin.org/') == ('httpbin.org', 'https://httpbin.org')
    assert get_hostinfo('https://example.org/path') == ('example.org', 'https://example.org')
    # keeping track of known URLs
    known_links = {'https://test.org'}
    assert is_known_link('https://test.org/1', known_links) is False
    assert is_known_link('https://test.org', known_links) is True
    assert is_known_link('http://test.org', known_links) is True
    assert is_known_link('http://test.org/', known_links) is True
    assert is_known_link('https://test.org/', known_links) is True


def test_external():
    '''test domain comparison'''
    assert is_external('https://github.com/', 'https://www.microsoft.com/') is True
    assert is_external('https://microsoft.com/', 'https://www.microsoft.com/', ignore_suffix=True) is False
    assert is_external('https://microsoft.com/', 'https://www.microsoft.com/', ignore_suffix=False) is False
    assert is_external('https://google.com/', 'https://www.google.co.uk/', ignore_suffix=True) is False
    assert is_external('https://google.com/', 'https://www.google.co.uk/', ignore_suffix=False) is True
    # malformed URLs
    assert is_external('h1234', 'https://www.google.co.uk/', ignore_suffix=True) is True


def test_extraction():
    '''test link comparison in HTML'''
    assert len(extract_links(None, 'https://test.com/', False)) == 0
    assert len(extract_links('', 'https://test.com/', False)) == 0
    # link known under another form
    pagecontent = '<html><a href="https://test.org/example"/><a href="https://test.org/example/&"/></html>'
    assert len(extract_links(pagecontent, 'https://test.org', False)) == 1
    # language
    pagecontent = '<html><a href="https://test.com/example" hreflang="de-DE"/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False)) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', True)) == 0
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='en')) == 0
    pagecontent = '<html><a href=https://test.com/example hreflang=de-DE/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 1
    # x-default
    pagecontent = '<html><a href="https://test.com/example" hreflang="x-default"/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='en')) == 1
    # language + content
    pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="https://test.com/example2"/><a href="https://test.com/example2 ADDITIONAL"/></html>'
    links = extract_links(pagecontent, 'https://test.com/', False)
    assert sorted(links) == ['https://test.com/example', 'https://test.com/example2']
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 2
    pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="https://test.com/page/2"/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False, with_nav=False)) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', False, with_nav=True)) == 2
    pagecontent = "<html><head><title>Links</title></head><body><a href='/links/2/0'>0</a> <a href='/links/2/1'>1</a> </body></html>"
    links = extract_links(pagecontent, 'https://httpbin.org', False, with_nav=True)
    assert sorted(links) == ['https://httpbin.org/links/2/0', 'https://httpbin.org/links/2/1']
    # links undeveloped by CMS
    pagecontent = '<html><a href="{privacy}" target="_privacy">{privacy-link}</a></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False)) == 0
    assert len(extract_links(pagecontent, 'https://test.com/', True)) == 0
    # links without quotes
    pagecontent = '<html><a href=/contact>Link</a></html>'
    assert extract_links(pagecontent, 'https://test.com/', False) == {'https://test.com/contact'}
    assert extract_links(pagecontent, 'https://test.com/', True) == set()
    pagecontent = '<html><a href=/contact attribute=value>Link</a></html>'
    assert extract_links(pagecontent, 'https://test.com/', False) == {'https://test.com/contact'}


def test_cli():
    '''test the command-line interface'''
    testargs = ['', '-i', 'input.txt', '-d', 'discardedfile.txt', '--outputfile', 'output.txt', '-v', '--language', 'en']
    with patch.object(sys, 'argv', testargs):
        args = cli.parse_args(testargs)
    assert args.inputfile == 'input.txt'
    assert args.discardedfile == 'discardedfile.txt'
    assert args.outputfile == 'output.txt'
    assert args.verbose is True
    assert args.language == 'en'
    assert os.system('courlan --help') == 0  # exit status

    # testfile
    inputfile = os.path.join(RESOURCES_DIR, 'input.txt')
    os_handle, temp_outputfile = tempfile.mkstemp(suffix='.txt', text=True)
    env = os.environ.copy()
    # Force encoding to utf-8 for Windows (seem to be a problem only in GitHub Actions)
    if os.name == 'nt':
        env['PYTHONIOENCODING'] = 'utf-8'
        courlan_bin = os.path.join(sys.prefix, "Scripts", "courlan")
    else:
        courlan_bin = 'courlan'
    # test for Windows and the rest
    assert subprocess.run([courlan_bin, '-i', inputfile, '-o', temp_outputfile], env=env).returncode == 0

    # tests without Windows
    if os.name != 'nt':
        # dry runs (writes to /tmp/)
        testargs = ['', '-i', inputfile, '-d', '/tmp/tralala1.txt', '-o', temp_outputfile, '--language', 'en', '--strict']
        f = io.StringIO()
        with patch.object(sys, 'argv', testargs):
            args = cli.parse_args(testargs)
        with redirect_stdout(f):
            cli.process_args(args)
        assert len(f.getvalue()) == 0
        testargs = ['', '-i', inputfile, '-o', '/tmp/tralala.txt', '--sample']
        with patch.object(sys, 'argv', testargs):
            args = cli.parse_args(testargs)
        with redirect_stdout(f):
            cli.process_args(args)
        assert len(f.getvalue()) == 0
    # delete temporary output file
    try:
        os.remove(temp_outputfile)
    except PermissionError:
        print("couldn't delete temp file")


def test_sample():
    '''test URL sampling'''
    assert not list(
        sample_urls(['http://test.org/test1', 'http://test.org/test2'], 0)
    )

    # assert len(sample_urls(['http://test.org/test1', 'http://test.org/test2'], 1)) == 1
    mylist = ['http://t.o/t1', 'http://test.org/test1', 'http://test.org/test2', 'http://test2.org/test2']
    assert len(sample_urls(mylist, 1, verbose=True)) == 2
    assert not sample_urls(mylist, 1, exclude_min=10, verbose=True)
    assert len(sample_urls(mylist, 1, exclude_max=1, verbose=True)) == 1

    test_urls = [f'https://test.org/{str(a)}' for a in range(1000)]
    example_urls = [f'https://www.example.org/{str(a)}' for a in range(100)]
    other_urls = [f'https://www.other.org/{str(a)}' for a in range(10000)]
    urls = test_urls + example_urls + other_urls
    sample = sample_urls(urls, 10)
    assert len([u for u in sample if 'test.org' in u]) == 10
    assert len([u for u in sample if 'example.org' in u]) == 10
    assert len([u for u in sample if 'other.org' in u]) == 10
    sample = sample_urls(urls, 150)
    assert len([u for u in sample if 'test.org' in u]) == 150
    assert len([u for u in sample if 'example.org' in u]) == 100
    assert len([u for u in sample if 'other.org' in u]) == 150


def test_examples():
    '''test README examples'''
    assert check_url('https://github.com/adbar/courlan') == ('https://github.com/adbar/courlan', 'github.com')
    assert check_url('https://httpbin.org/redirect-to?url=http%3A%2F%2Fexample.org', strict=True) == ('https://httpbin.org/redirect-to', 'httpbin.org')
    assert clean_url('HTTPS://WWW.DWDS.DE:80/') == 'https://www.dwds.de'
    assert validate_url('http://1234') == (False, None)
    assert validate_url('http://www.example.org/')[0] is True
    assert normalize_url('http://test.net/foo.html?utm_source=twitter&post=abc&page=2#fragment', strict=True) == 'http://test.net/foo.html?page=2&post=abc'



"""
Functions devoted to requests over the WWW.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import logging
import urllib3

RETRY_STRATEGY = urllib3.util.Retry(
    total=5,
    redirect=5,
    raise_on_redirect=False,
)
HTTP_POOL = urllib3.PoolManager(retries=RETRY_STRATEGY)

ACCEPTABLE_CODES = {200, 300, 301, 302, 303, 304, 305, 306, 307, 308}


# Test redirects
def redirection_test(url):
    """ Test final URL to handle redirects
    Args:
        url: url to check

    Returns:
        The final URL seen.

    Raises:
        Nothing.
    """
    #headers.update({
    #    "User-Agent" : str(sample(settings.USER_AGENTS, 1)), # select a random user agent
    #})
    try:
        rhead = HTTP_POOL.request('HEAD', url)
    except Exception as err:
        logging.error('unknown: %s %s', url, err) # sys.exc_info()[0]
    else:
        # response
        if rhead.status in ACCEPTABLE_CODES:
            logging.debug('result found: %s %s', rhead.geturl(), rhead.status)
            return rhead.geturl()
    #else:
    logging.debug('no result found: %s', url)
    return None
